/*
 *  xgridloc: An X/GTK+ application to calculate the Maidenhead Grid
 *  Locator of a given position or vice-versa. Also to calculate
 *  bearing and distance to a remote position and plot the Great
 *  Circle arc to it on screen, using xplanet if installed.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */


/*  main.c
 *
 *  Main body of gridloc application
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "interface.h"
#include "support.h"
#include "callbacks.h"
#include "xgridloc.h"

char
  home_posn[30],  /* 'Home' position string */
  home_name[25];  /* 'Home' position name   */

/* Signal handler */
static void Signal_Handler( int signal );

/* xgridloc's top window */
GtkWidget *xgridloc_main_window;

/* Longitude and Latitude */
extern double
  home_lon, home_lat,
  remt_lon, remt_lat;


int
main( int argc, char *argv[] )
{
  /* Command line option returned by getopt() */
  int option;

  /* getopt() variables */
  extern char *optarg;
  extern int optind, opterr, optopt;

  /* New and old actions for sigaction() */
  struct sigaction sa_new, sa_old;

  /* Initialize new actions */
  sa_new.sa_handler = Signal_Handler;
  sigemptyset( &sa_new.sa_mask );
  sa_new.sa_flags = 0;

  /* Register function to handle signals */
  sigaction( SIGINT,  &sa_new, &sa_old );
  sigaction( SIGSEGV, &sa_new, 0 );
  sigaction( SIGFPE,  &sa_new, 0 );
  sigaction( SIGTERM, &sa_new, 0 );
  sigaction( SIGABRT, &sa_new, 0 );

  /* Process command line options */
  while( (option = getopt(argc, argv, "hv") ) != -1 )
	switch( option )
	{
	  case 'h' : /* Print usage and exit */
		Usage();
		return(0);

	  case 'v' : /* Print version */
		fprintf( stderr, "%s %s\n", PACKAGE, VERSION );
		return(0);

	  default: /* Print usage and exit */
		Usage();
		exit(-1);

	} /* End of switch( option ) */

  gtk_set_locale();
  gtk_init(&argc, &argv);

  xgridloc_main_window = create_main_window();

  /* Setup signal callbacks */
  {
	GtkWidget *entry;

	entry = lookup_widget(xgridloc_main_window, "home_lon_deg");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_home_pos_activate), "179");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "179");

	entry = lookup_widget(xgridloc_main_window, "home_lon_min");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_home_pos_activate), "59");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "59");

	entry = lookup_widget(xgridloc_main_window, "home_lon_sec");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_home_pos_activate), "59");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "59");

	entry = lookup_widget(xgridloc_main_window, "home_lat_deg");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_home_pos_activate), "89");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "89");

	entry = lookup_widget(xgridloc_main_window, "home_lat_min");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_home_pos_activate), "59");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "59");

	entry = lookup_widget(xgridloc_main_window, "home_lat_sec");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_home_pos_activate), "59");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "59");

	entry = lookup_widget(xgridloc_main_window, "remt_lon_deg");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_remt_pos_activate), "179");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "179");

	entry = lookup_widget(xgridloc_main_window, "remt_lon_min");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_remt_pos_activate), "59");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "59");

	entry = lookup_widget(xgridloc_main_window, "remt_lon_sec");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_remt_pos_activate), "59");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "59");

	entry = lookup_widget(xgridloc_main_window, "remt_lat_deg");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_remt_pos_activate), "89");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "89");

	entry = lookup_widget(xgridloc_main_window, "remt_lat_min");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_remt_pos_activate), "59");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "59");

	entry = lookup_widget(xgridloc_main_window, "remt_lat_sec");
	g_signal_connect ((gpointer) entry, "activate",
		G_CALLBACK (on_remt_pos_activate), "59");
	g_signal_connect ((gpointer) entry, "changed",
		G_CALLBACK (on_pos_changed), "59");
  }

  gtk_widget_show(xgridloc_main_window);
  gtk_idle_add( Load_Config, "" );

  gtk_main();

  return 0;
}

/*------------------------------------------------------------------------*/

/*  Load_Config()
 *
 *  Loads the xgridloc configuration file
 */

  gboolean
Load_Config( gpointer data )
{
  char
    *rc_fpath, /* File path to xgridlocrc */
    line[81];  /* Buffer for Load_Line   */

  /* Config file pointer */
  FILE *xgridlocrc;


  /* Setup file path to xgridlocrc */
  rc_fpath = malloc( strlen(getenv("HOME")) + strlen(".xgridlocrc") + 2 );
  if( rc_fpath == NULL )
  {
	Error_Dialog(
		"Failed to allocate memory for\n"
		"Configuration file name string.\n"
		"Please quit xgridloc and correct" );
	return( FALSE );
  }
  sprintf( rc_fpath, "%s/%s", getenv("HOME"), ".xgridlocrc" );

  /* Open ~./gridlocrc file */
  xgridlocrc = fopen( rc_fpath, "r" );
  if( xgridlocrc == NULL )
  {
	Error_Dialog(
		"Error opening configuration file\n"
		"Please quit xgridloc and correct" );
	return( FALSE );
  }
  free(rc_fpath);

  /*** Read runtime configuration data ***/
  /* Read 'Home' position, abort on EOF */
  if( Load_Line(line, xgridlocrc, "Home Position" ) != SUCCESS )
	return( FALSE );
  strncpy( home_posn, line, 29 );
  home_posn[29] = '\0';

  /* Read 'Home' position name, abort if EOF */
  if( Load_Line(line, xgridlocrc, "Home Position Name" ) != SUCCESS )
	return( FALSE );
  strncpy( home_name, line, 24 );
  home_name[24] = '\0';

  fclose( xgridlocrc );

  return( FALSE );

} /* End of Load_Config */

/*------------------------------------------------------------------*/

/*  Load_Line()
 *
 *  Loads a line from a file, aborts on failure. Lines beginning
 *  with a '#' are ignored as comments. At the end of file EOF is
 *  returned. Lines assumed maximum 80 characters long.
 */

  int
Load_Line( char *buff, FILE *pfile, char *messg )
{
  int
	num_chr, /* Number of characters read, excluding lf/cr */
	chr;     /* Character read by getc() */
  char err_msg[128];

  /* Prepare error message */
  snprintf( err_msg, 128,
	  "Error reading %s\n"
	  "Premature EOF (End Of File)", messg );

  /* Clear buffer at start */
  buff[0] = '\0';
  num_chr = 0;

  /* Get next character, return error if chr = EOF */
  if( (chr = fgetc(pfile)) == EOF )
  {
	fprintf( stderr, "xgridloc: %s\n", err_msg );
	Error_Dialog( err_msg );
	return( EOF );
  }

  /* Ignore commented lines and eol/cr */
  while(
	  (chr == '#') ||
	  (chr == ' ') ||
	  (chr == HT ) ||
	  (chr == CR ) ||
	  (chr == LF ) )
  {
	/* Go to the end of line (look for LF or CR) */
	while( (chr != CR) && (chr != LF) )
	  /* Get next character, return error if chr = EOF */
	  if( (chr = fgetc(pfile)) == EOF )
	  {
		fprintf( stderr, "xgridloc: %s\n", err_msg );
		Error_Dialog( err_msg );
		return( EOF );
	  }

	/* Dump any CR/LF remaining */
	while( (chr == CR) || (chr == LF) )
	  /* Get next character, return error if chr = EOF */
	  if( (chr = fgetc(pfile)) == EOF )
	  {
		fprintf( stderr, "xgridloc: %s\n", err_msg );
		Error_Dialog( err_msg );
		return( EOF );
	  }

  } /* End of while( (chr == '#') || ... */

  /* Continue reading characters from file till
   * number of characters = 80 or EOF or CR/LF */
  while( num_chr < 80 )
  {
	/* If LF/CR reached before filling buffer, return line */
	if( (chr == LF) || (chr == CR) ) break;

	/* Enter new character to line buffer */
	buff[num_chr++] = chr;

	/* Get next character */
	if( (chr = fgetc(pfile)) == EOF )
	{
	  /* Terminate buffer as a string if chr = EOF */
	  buff[num_chr] = '\0';
	  return( SUCCESS );
	}

	/* Abort if end of line not reached at 80 char. */
	if( (num_chr == 80) && (chr != LF) && (chr != CR) )
	{
	  /* Terminate buffer as a string */
	  buff[num_chr] = '\0';
	  snprintf( err_msg, 128,
		  "Error reading %s\n"
		  "Line longer than 80 characters", messg );
	  fprintf( stderr, "xgridloc: %s\n%s\n", err_msg, buff );
	  Error_Dialog( err_msg );
	  return( ERROR );
	}

  } /* End of while( num_chr < max_chr ) */

  /* Terminate buffer as a string */
  buff[num_chr] = '\0';

  return( SUCCESS );

} /* End of Load_Line() */

/*------------------------------------------------------------------*/

/*  Sig_Handler()
 *
 *  Signal Action Handler function
 */

static void Signal_Handler( int signal )
{
  Cleanup();

  fprintf( stderr, "\n" );
  switch( signal )
  {
    case SIGINT  : fprintf( stderr,  "gridloc: Exiting via User Interrupt" );
		   exit(0);
    case SIGSEGV : fprintf( stderr,  "gridloc: Segmentation Fault" );
		   break;
    case SIGFPE  : fprintf( stderr,  "gridloc: Floating Point Exception" );
		   break;
    case SIGABRT : fprintf( stderr,  "gridloc: Abort Signal received" );
		   break;
    case SIGTERM : fprintf( stderr,  "gridlock: Termination Request received" );
  }

  exit( -1 );

} /* End of Sig_Handler() */

/*------------------------------------------------------------------------*/

/*  Cleanup()
 *
 *  Cleans up before quitting
 */

void
Cleanup(void)
{
  /* Delete xplanet files */
  if( isFlagSet(XPLANET_ENABLED) )
  {
    unlink( CONFIG_FILE );
    unlink( MARKER_FILE );
    unlink( ARC_FILE );
	system( "pkill xplanet" );
  }
}

/*------------------------------------------------------------------------*/

/* Functions for testing and setting/clearing flags */

/* An int variable holding the single-bit flags */
static int Flags = 0;

int
isFlagSet(int flag)
{
  return(Flags & flag);
}

int
isFlagClear(int flag)
{
  return(~Flags & flag);
}

void
SetFlag(int flag)
{
  Flags |= flag;
}

void
ClearFlag(int flag)
{
  Flags &= ~flag;
}

void
ToggleFlag(int flag)
{
  Flags ^= flag;
}

/*------------------------------------------------------------------*/

/*  Usage()
 *
 *  Prints usage information
 */

  void
Usage( void )
{
  fprintf( stderr, "%s\n", "Usage: xgridloc [-hv]" );

  fprintf( stderr, "%s\n",
	  "       -h: Print this usage information and exit");

  fprintf( stderr, "%s\n",
	  "       -v: Print version number and exit");

} /* End of Usage() */

/*------------------------------------------------------------------------*/

